"""
The data types for the Simba interpreter.
Many data types are declared as aliases for the benefit of being able to call isinstance() to determine type (Maybe this is a bad idea).
"""
from re import split
from simba.simba_exceptions import UnresolvedSymbolError, ImmutableBindingException, SimbaSyntaxError, SimbaException
import pyrsistent as p

# Atomic Data Types
class Symbol:
    def __init__(self, string):
        split_str = split('/', string)
        if string == '/': self.name = '/'; self.namespace = None
        elif len(split_str) > 2: raise SimbaSyntaxError(f"A symbol can only have one namespace qualifier (symbol `{string}`).")
        elif len(split_str) == 2:
            if split_str[0] == '' or split_str[1] == '': raise SimbaSyntaxError("None of the two members of the symbol should be empty.", string, None, None)
            self.namespace = split_str[0]
            self.name = split_str[1]
        elif len(split_str) == 1:
            if split_str[0] == '': raise SimbaSyntaxError("A symbol cannot be empty.")
            self.namespace = None
            self.name = split_str[0]
    def __str__(self):
        if self.namespace: return '/'.join([self.namespace, self.name])
        else: return self.name
    def __repr__(self):
        return str(self)
    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Symbol): return False
        if __o.name == self.name and __o.namespace == self.namespace: return True
        return False

Environment = dict
class Keyword(str): pass

# Symbolic Data Types

class SymbolicExpression():
    """A symbolic expression contains: symbol, positional arguments, relational arguments"""
    def __init__(self, *args, **kwargs):
        self.positional = []
        self.relational = {**kwargs}
        _skip = False
        for i, e in enumerate(args):
            if _skip: _skip = False; continue
            if not isinstance(e, Keyword):
                self.positional.append(e)
            else:
                self.relational[str(e)] = args[i+1]
                _skip = True

    def append(self, e):
        self.positional.append(e)

    def __repr__(self):
        return '<SymbolicExpression>(' + str(self.positional) + str(self.relational) + ')'

    def __contains__(self, item):
        """This method overrides the Python `in` operator:
        - If `item` is a Keyword, look in the relational arguments
        - else, look in the positional arguments"""
        if isinstance(item, Keyword) or isinstance(item, str):
            return str(item) in self.relational
        else:
            return item in self.positional

    def __getitem__(self, __slice):
        if isinstance(__slice, str):
            return self.relational[__slice]
        else:
            return self.positional[__slice]

    def __eq__(self, __o):
        if not isinstance(__o, SymbolicExpression): return False
        return  len(self) == len(__o) \
            and all(value == __o[i] for i, value in enumerate(self.positional)) \
            and all(key in __o and self[key] == __o[key] for key in self.relational if key != 'meta')

    def __len__(self):
        return len(self.positional) + len(self.relational)

    def __add__(self, sexp):
        if isinstance(sexp, SymbolicExpression):
            self.positional = self.positional + sexp.positional
            self.relational = self.relational | sexp.relational
            return self
        # elif isinstance(sexp, tuple):
        #     self.positional = self.positional + SymbolicExpression(sexp).positional
        #     return self
        else:
            raise SimbaException(f"Tried to concat SymbolicExpression with unsupported type {type(sexp)}")

# Collection Data Types

List = p.PList
Vector = p.PVector
Map = p.PMap

class Protocol():
    """This is the protocol type"""
    pass

class Environment():
    """
    Just like a scheme environment. Contains a mapping of symbols to namespaces and an outer environment.
    Corresponds to the scope of a `let` form. Importantly, the bindings created are immutable.
    """
    def __init__(self, outer, names = None):
        self.outer = outer
        self.names = {} if names is None else names
        self.ns = outer.ns
    def get(self, symbol):
        """Looks for a binding, first in current env, and then in outer, and so on.
        Return the value bound to the given symbol."""
        n = symbol.name
        # 1. if the symbol is namespace-qualified, look in the appropriate namespace
        if symbol.namespace is not None:
            self.ns.get(symbol)
        # 2. else look in the local names
        if symbol.name in self.names:
            return self.names[n]
        # 3. if not found, look in outer envs
        else:
            return self.outer.get(symbol)
        # NB: You only want to look for referred names at the ns level
    def set(self, symbol, val):
        if symbol.namespace: raise SimbaException(f"Illegal definition of the namespaced symbol `{symbol.namespace}/{symbol.name}` in a let statement.")
        self.names[symbol.name] = val

from simba.base_functions import repl_env

class Namespace():
    """Just like an environment except that it has a dictionary of required and included namespaces.
    A name that equals to the empty string denotes that the env is an empty shell and has to be mutated."""
    def __init__(self, name = "", names = None):
        self.name = name
        self.names = names if names else repl_env.copy()
        self.namespaces = {}
        self.referred = []
        self.ns = self
    def set(self, symbol, value):
        """Change or add a binding. The binding has to be in the current namespace."""
        if symbol.namespace is not None and symbol.namespace != self.name:
            raise ImmutableBindingException(f"Cannot mutate the binding `{symbol}` from the current namespace {self.name}.")
        self.names[symbol.name] = value
    def get(self, symbol):
        """Looks for a binding in namespace and . Returns the env where the match is made first.
        If the binding is namespace-qualified and doesn't exist in that ns, throws an error.
        Else, if the binding doesn't exist, raise an error"""
        n = symbol.name
        ns = symbol.namespace
        # 1. look in external namespaces if the symbol is ns-qualified
        if ns is not None and ns != self.name:
            try:
                # makes sure the binding exists in the external namespace, 
                # otherwise throwing an error is the expected behavior
                return self.namespaces[ns].names[n]
            except KeyError:
                raise UnresolvedSymbolError(f"\n\tThe symbol `{symbol}` could not be found.\n\tMake sure the namespace is correctly imported and the symbol exists in that namespace.")
        # 2. else look in current ns
        # print(self.name)
        if n in self.names:
            # print(f"{symbol} is in names")
            return self.names[n]
        # 3. lastly, look in the referred namespaces
        for refd_ns in self.referred:
            if n in refd_ns.names:
                return refd_ns.names[n]
        raise UnresolvedSymbolError(symbol)
    def require_ns(self, ns):
        self.namespaces[ns.name] = ns
    def include_ns(self, ns):
        self.require_ns(ns)
        self.referred.append(ns)


import simba
import simplest_test_suite

## Running the Python tests ##

t = simplest_test_suite.simple_tests
tests_done = 0
for i in range(len(t)):
    if (i % 2) == 0:
        res = simba.eval_str(t[i])
        # try:
        print(t[i], '=>', res)
        assert(res == t[i+1])
        # except:
        #     print(f"Error: {simba.print_sexp(t[i])} == {res} != {t[i+1]}")
        tests_done = tests_done + 1

print(f"{tests_done} tests successfully passed.")


# def run(file):
#     input  = antlr4.FileStream(file, encoding='utf-8')
#     with open(file, 'r') as f:
#         code       = f.read_sexp()
#         unindented = reader.addIndentationTokens(code)
#         ds         = reader.read_sexp(antlr4.InputStream(unindented), {})
#         result     = eval_sexp(ds, {})
#         print(result)

            # print(rep(antlr4.StdinStream(encoding='utf-8')))

# lexer = simbaLexer(StdinStream())
# stream = CommonTokenStream(lexer)
# parser = simbaParser(stream)
# tree = parser.hi()
# printer = simbaPrintListener()
# walker = ParseTreeWalker()
# walker.walk(printer, tree)

# input  = FileStream(filename, encoding='utf-8') if filename else StdinStream(encoding='utf-8')

# def eval_sexp(sexp, env):
#     """
#     eval_exp evaluates a Symbolic Expression within a context.
#     The context is the 'environment' a mapping of names to namespaces,
#     and the namespace in which the execution is working at the moment.
#     By default, execution starts in the `None` namespace, which corresponds to the standard library.
#     """
#     if (isinstance(sexp, Symbol)): # if symbol evaluate to its binding
#         try: return env.get(sexp)
#         except KeyError: raise UnresolvedSymbolError(sexp)
#     # elif map
#     # elif vector
#     elif (isinstance(sexp, SymbolicExpression)): # apply the symbolic expression
#         if len(sexp.positional) == 0:
#             return None
#         ## Special Forms: ##
#         head = sexp.positional[0]
#         if "def" == head.name:
#             ## def adds a binding (from unevaluated first arg to evaled second arg) in the environment ##
#             a1, a2 = sexp.positional[1], sexp.positional[2]
#             res = eval_sexp(a2, env)
#             return env.set(a1, res)
#         elif "quote" == head.name:
#             if len(sexp.positional) != 2: raise Exception("quote expected 1 argument")
#             return sexp.positional[1]
#         elif "if" == head.name:
#             ## If may need some more work: what is truthy? what if fewer then 3 args? ##
#             if (len(sexp.positional) < 3):
#                 raise SyntaxError("Too few arguments to if")
#             if (len(sexp.positional) > 4):
#                 raise SyntaxError("Too many arguments to if")
#             cond, then = sexp.positional[1:3]
#             if cond:
#                 return eval_sexp(then, env)
#             else:
#                 if (len(sexp.positional) == 4):
#                     _else = sexp.positional[3]
#                     return eval_sexp(_else, env)
#                 else: return None
#         elif "fn" == head.name:
#             ## Anonymous function ##
#             ## returns a new closure ##
#             argVector, body = sexp.positional[1], sexp.positional[2:]
#             # what the closure should do is:
#             # - gather all the free variables
#             # - point to the local bindings in `bindings`
#             bindings = env # this is temporary: 
#             # for now, the entire enclosing scope in copied
#             # in the future, the bindings will be only the ones references in the closure
#             return Function(argVector, body, bindings)
#         elif "do" == head.name:
#             last = None
#             for exp in sexp.positional[1:]:
#                 last = eval_sexp(exp, env)
#             return last
#         elif "comment" == head.name:
#             return None
#         elif "let" == head.name:
#             ## creates a new environment and binds the values to it successively (in order) ##
#             vec, body = sexp.positional[1], sexp.positional[2:]
#             let_env = SimbaEnvironment(outer = env)
#             for i in range(0, len(vec), 2):
#                 let_env.set(vec[i], eval_sexp(vec[i+1], let_env))
#             ## evaluate the body in an implicit do ##
#             last = None
#             for exp in body:
#                 last = eval_sexp(exp, let_env)
#             return last
#         # elif type(head) == function:
#         #     ## Call a native function ##
#         #     return head(*sexp.positional[1:], **sexp.relational)
#         elif "ns" == head.name:
#             # ## If the namespace is 
#             # ## add a ns to the environment, and "jump" into it for evaluation ##
#             # if len(sexp.positional) != 2:
#             #     raise SyntaxError("The `namespace` form should have one argument.")
#             # # 1. Initialize the namespace and add it to the namespaces
#             # new_ns = sexp.positional[1]
#             # namespaces[new_ns] = SimbaEnvironment()
#             # # 2. Jump into the namespace to start evaluation

#             # print(env)
#             last = None
#             for exp in sexp.positional[2:]:
#                 last = eval_sexp(exp, env)
#             return last
#         else:
#             ## Non-Special Forms ##
#             ## evaluate the args ##
#             evaledPArgs = [eval_sexp(e, env) for e in sexp.positional]
#             evaledRArgs = {eval_sexp(key, env): eval_sexp(value, env) for key, value in sexp.relational}
#             ## apply the head ##
#             return evaledPArgs[0](*evaledPArgs[1:], **evaledRArgs)
#     else: # if the value is an atomic data type, cannot evaluate further
#         return sexp

# def eval_str(string):
#     exec_environment = SimbaEnvironment(names = repl_env)
#     readerObj = _reader.SexpReader(string)
#     result = None
#     while readerObj.position < len(readerObj.tokens):
#         e = readerObj.read_form()
#         result = eval_sexp(e, exec_environment)
#     return result

# elif "for" == head.name:
#     res = None
#     for 
# elif "while" == head.name:
#     cond = sexp.positional[1]
#     body = cond.positional[2:]
#     while ns_eval_sexp(cond, env, loaded_namespaces):
#         ns_eval_sexp(body, env, ns_eval_sexp)
# elif "loop" == head.name:
#     # ensure recur is at tail position
#     bindings = sexp[1] # hopefully the first argument is a vector
#     # ensure that the recur happens at the tails of the expr
#     contents = sexp[2:]
#     # val = 
#     while true:
#         # when you encounter recur, call next
#         # at the end, call break

# print(isinstance(pyrsistent.pvector([1, 2, 3]), collections.abc.Sequence))

a = {b'op': b'eval', b'code': b'*ns*', b'id': b'1'}
print(type(b'op'))

print(isinstance(b'op', bytes))
# print(a.items())

# https://python-ast-explorer.com/
# https://github.com/octoml/synr


#! python3
import sys, traceback
from types import ModuleType
from typing import Tuple, Union
from importlib.resources import read_text

import argparse
from prompt_toolkit.shortcuts.prompt import PromptSession
from prompt_toolkit.history import FileHistory

import pyrsistent as p

# import antlr4
import simba.sexprs_reader_printer as sexprs_reader_printer
# import src.simba.dotexprs_reader_printer
from simba.simba_types import *
import simba.helpers as helpers
import simba.sb as sb

# _reader    = sexprs_reader_printer.SexpReader
# _read_form  = sexprs_reader_printer.read_str
# _print_ast = sexprs_reader_printer.to_string
_reader = sexprs_reader_printer

default_syntax = sexprs_reader_printer

# def read_sexp(string, syntax = default_syntax): # returns a SymbolicExpression
#     # read the frontmatter
#     return syntax.read_str(string)

def read_string(string):
    """Returns the first object from the string."""
    return _reader.SexpReader(string).read_form()

def print_sexp(sexp, syntax = default_syntax, lb=False) -> str:
    return syntax.to_string(sexp, lb=lb)

from simba.simba_exceptions import SimbaException, MultipleDispatchException, FailingGuardException, MultiMethodException, SimbaSyntaxError

def bind_fn_args(fn_args:tuple, p_args: tuple, r_args: dict) -> dict:
    names = {}
    for i, arg in enumerate(fn_args):
        if arg.name == '&':
            names[fn_args[i+1].name] = p_args[i:]
            break
        else:
            names[arg.name] = p_args[i]
    return names

class Function:
    """Functions are anonymous, but may contain names in their metadata.
    Any Function can be promoted to a multimethod with the `register` method.
    """
    def __init__(self, params, ast, bindings, _r_args = None):
        # positonalParams = pass
        # namedParams = pass
        r_args = _r_args if _r_args is not None else {}
        self.args = params
        self.ast = ast
        self.guard = None if 'when' not in r_args else r_args['when']
        self.bindings = bindings # the bindings should hold the bindings from the args and closure
        self.macro = r_args['macro'] if 'macro' in r_args else False
        self.method_table = []
        if 'meta' in r_args:
            self.meta = r_args['meta']
        else:
            self.meta = {}
        for k in ['macro', 'meta']:
            if k in r_args: r_args.pop(k)
        self.meta = self.meta | r_args
    def register(self, fn):
        return Multi(self, **self.meta, macro = self.macro).register(fn) # add meta
    def __call__(self, *e_p_args, **e_r_args):
        res =  None
        env = Environment(
                outer=self.bindings,
                names=bind_fn_args(self.args, e_p_args, e_r_args)
            )
        for e in self.ast:
            res = eval_sexp(e, env = env)
        return res

def check_guard(guard:Union[Function,None], p_args:tuple, r_args:dict):
    if guard:
        return guard(*p_args, **r_args)
    else:
        return True

def signature_match(signature:tuple, p_args:tuple, r_args:dict, guard: Union[Function,None] = None) -> bool:
    if Symbol('&') in signature:
        return check_guard(guard, p_args, r_args) and True
    else:
        # TODO: for now the signature length does not take into account the keyword arguments or type signatures
        return len(p_args) == len(signature) and check_guard(guard, p_args, r_args)

class Multi:
    """All functions can be promoted to multimethods.
    """
    def __init__(self, fn:Function, **meta):
        self.method_table: List[Tuple[Tuple, Function, Union[Function, None]]] = []
        self.meta = meta
        if 'macro' in self.meta:
            self.macro = self.meta['macro']
            self.meta.pop('macro')
        else: 
            self.macro = False
        self.register(fn)
    def register(self, fn: Function):
        self.method_table.append((fn.args, fn, fn.guard))
        return self
    def clear(self):
        self.method_table = []
    def signatures(self):
        r = []
        for t in self.method_table:
            r.append((t[0], t[2]))
        return r
    def __len__(self):
        return len(self.method_table)
    def __call__(self, *p_args, **r_args):
        for tup in self.method_table:
            if signature_match(tup[0], p_args, r_args, guard = tup[2]):
                return tup[1](*p_args, **r_args)
        n = self.meta['name'] if 'name' in self.meta else ""
        raise MultipleDispatchException(f"No matching signature for multimethod {n}")

class MultiFn():
    """A multimethod. Dispatch is made according to the result of the dispatching function."""
    def __init__(self, name, dispatch_fn):
        self.name = name
        self.dispatch_fn = dispatch_fn
        self.method_table = {} # a mapping of dispatch-value to functions
    def __len__(self):
        return len(self.method_table)
    def register(self, fn, dispatch_value):
        if dispatch_value in self.method_table:
            raise MultiMethodException(f"The dispatch value {print_sexp(dispatch_value)} is already registered for the multimethod {self.name}.")
        self.method_table[dispatch_value] = fn
    def methods(self):
        """Returns a list of the registered dispatch values on the multimethod."""
        return p.pvector(self.method_table.keys())
    # def __call__(self):
    #     dispatch_val = self.dispatch_fn()
    #     # if val not in dict dispatch to the default fn?
    #     return self.method_table[dispatch_val]()

def convert(string, _from, to):
    return to.print_sexp(_from.read_sexp(string))

# def eval_form(env, form, print_result = False):
#     result = eval_sexp(form, env)
#     if print_result: print(print_sexp(result))

def simba_repl(multi = False):
    """Command-line repl that executes everything in the base namespace by default."""
    eof = False
    prompt_session = PromptSession(FileHistory('~/.simba_history'))
    # load the standard library
    _env = Namespace()
    eval_sexp(SymbolicExpression(Symbol('ns'), Symbol('repl')), _env)
    while not eof:
        try:
            line = prompt_session.prompt('>>> ', multiline=multi)
            readerObj = _reader.SexpReader(line)
            while readerObj.position < len(readerObj.tokens):
                e = readerObj.read_form()
                res = eval_sexp(e, _env)
                print(print_sexp(res))
        except EOFError:
            eof = True
        except KeyboardInterrupt:
            print('Exit with Ctrl-D')
            eof = True
        except Exception as e:
            print("".join(traceback.format_exception(*sys.exc_info())))

def quasiquote(ast):
    """The quasiquote expansion algorithm"""
    if isinstance(ast, Symbol) or isinstance(ast, Map): 
        return SymbolicExpression(Symbol('quote'), ast)
    # if isinstance(ast, Keyword):
    #     return SymbolicExpression(Symbol('quote'), )
    elif isinstance(ast, SymbolicExpression) and ast[0] == Symbol("unquote"):
        return ast[1]
    elif isinstance(ast, SymbolicExpression):
        relational = []
        for k in ast.relational:
            relational.append(Keyword(k)); relational.append(ast.relational[k])
        res = SymbolicExpression()
        for e in helpers.reverse(ast.positional):
            if isinstance(e, SymbolicExpression) and e[0] == Symbol("splice-unquote"):
                res = SymbolicExpression(Symbol('concat'), e[1], res)
            else:
                res = SymbolicExpression(Symbol('prepend-sexp'), quasiquote(e), res)
        # TODO: DOES THIS ALGORITHM PERFORM MACROEXPANSION ON KEYWORD ARGUMENTS??
        res = SymbolicExpression(Symbol('concat'), res, relational)
        return res
    else:
        return ast

def is_macro_call(ast, env):
    if isinstance(ast, SymbolicExpression) and len(ast.positional) != 0 and isinstance(ast[0], Symbol):
        try:
            refers_to = env.get(ast[0])
        except:
            refers_to = None
        if isinstance(refers_to, Function) and refers_to.macro == True:
            return True
    return False

def macroexpand(ast, env):
    """This is the most basic macroexpansion algorithm possible:
    While the head of the expression resolves to a macro, the corresponding
    macro function is called on its **unevaluated** arguments.

    Further improvements to this algorithm could be automatically namespaced
    symbols or autogensyms.
    """
    while is_macro_call(ast, env):
        macro_fn = env.get(ast[0])
        p_args = ast[1:]
        r_args = ast.relational
        ast = macro_fn(*p_args, **r_args)
    return ast

loaded_ns = {}

def require(ns_names, env, include = False):
    """Side effecting function that requires a namespace such that it is accessible in the current namespace (if namespace-qualified)."""
    global loaded_ns
    for n in ns_names:
        if n not in loaded_ns:
            env_cache = env
            # creates a new environment for the new namespace
            ns_env = Namespace(name = n)
            # evaluates the content of that namespace in the new environment
            # here the ns form will change the *ns*
            for exp in helpers.return_ns(n, program_ast):
                eval_sexp(exp, ns_env)
            # add the new namespace environment to the list of loaded namespaces
            loaded_ns[n] = ns_env
            # add the environment to the list of contextual namespaces
            if include: env.include_ns(loaded_ns[n])
            else:       env.require_ns(loaded_ns[n])
            # here we must change the *ns* back
            env = env_cache
        else:
            if include: env.include_ns(loaded_ns[n])
            else:       env.require_ns(loaded_ns[n])

def eval_primitive(sexp, env):
    """This function is charged of evaluating anything that is not a SymbolicExpression.
    In other words it just returns primitive data structures like Symbols, Vectors, etc...
    This distinction is made so that it is feasible to implement Tail Call Optimization
    on Symbolic Expression evaluation.
    """
    if isinstance(sexp, Symbol): # if symbol evaluate to its binding
        if sexp.name == "*ns*":
            return env.ns
        elif sexp.name == "*env*":
            return env
        if sexp.namespace:
            if sexp.namespace == "py":
                return eval(sexp.name)
            try:
                resolution = env.get(Symbol(sexp.namespace))
                if isinstance(resolution , ModuleType):
                    return getattr(resolution, sexp.name)
                # here it would be cleaner if a symba ns was the same as a python module, just a var in global scope
            except UnresolvedSymbolError:
                pass # do nothing
        try: return env.get(sexp)
        except KeyError: raise UnresolvedSymbolError(sexp)
    elif isinstance(sexp, Vector):
        return p.pvector([eval_sexp(e, env) for e in sexp])
    elif isinstance(sexp, Map):
        return p.pmap({eval_sexp(key, env): eval_sexp(sexp[key], env) for key in sexp})
    else: # if the value is an atomic data type, cannot evaluate further
        return sexp

def eval_sexp(sexp, env):
    """`eval_exp` evaluates a Symbolic Expression within a context.
    The context is the 'environment' a mapping of names to namespaces,
    and the namespace in which the execution is working at the moment.
    By default, execution starts in the `None` namespace, which corresponds to the standard library.
    """
    global loaded_ns
    while True:
        # print(sexp)
        if not isinstance(sexp, SymbolicExpression):
            return eval_primitive(sexp, env)

        # macroexpansion step
        sexp = macroexpand(sexp, env)

        # We need to check for primitive datatypes before and after macroexpansion unfortunately
        if not isinstance(sexp, SymbolicExpression):
            return eval_primitive(sexp, env)

        if len(sexp.positional) == 0:
            return None
        ## Special Forms: ##
        if isinstance(head := sexp.positional[0], SymbolicExpression):
            head = sexp.positional[0] = eval_sexp(head, env)
        head = sexp.positional[0]
        is_s = isinstance(head, Symbol)
        if is_s and "def" == head.name:
            if len(sexp.positional) < 3:
                raise SimbaSyntaxError(f"\n\t`def` statement has invalid argument pattern in:\n\t{print_sexp(sexp)}")
            _name = sexp.positional[1]
            _val = sexp.positional[2]
            _res = eval_sexp(_val, env)
            return env.ns.set(_name, _res)
        elif is_s and "quote" == head.name:
            if len(sexp.positional) != 2: raise SimbaSyntaxError(f"\n\t`quote` expected 1 argument in:\n\t{print_sexp(sexp)}")
            return sexp.positional[1]
        elif is_s and "if" == head.name:
            if len(sexp.positional) < 3:
                raise SimbaSyntaxError(f"\n\tToo few arguments to if in:\n\t{print_sexp(sexp)}")
            if len(sexp.positional) > 4:
                raise SimbaSyntaxError(f"\n\tToo many arguments to if in:\n\t{print_sexp(sexp)}")
            cond, then = sexp.positional[1:3]
            evaled_cond = eval_sexp(cond, env)
            if evaled_cond:
                sexp = then 
                continue # TCO
            else:
                if (len(sexp.positional) == 4):
                    _else = sexp.positional[3]
                    sexp = _else 
                    continue # TCO
                else: return None
        elif is_s and "fn" == head.name:
            arg1 = sexp.positional[1]
            if isinstance(arg1, Vector):
                # anonymous function (single method declaration)
                argVector, body = sexp.positional[1], sexp.positional[2:]
                # what the closure should do is:
                # - gather all the free variables
                # - point to the local bindings in `bindings`
                bindings = env # this is temporary:
                # for now, the entire enclosing scope in referenced. This may impair garbage collection too much.
                # in the future, the bindings will be only the ones references in the closure
                return Function(argVector, body, bindings, sexp.relational)
            elif isinstance(arg1, SymbolicExpression):
                # anonymous multimethod (with multiple method declarations)
                fn = Function(sexp[1][0], sexp[1][1:], env, sexp.relational)
                for exp in sexp[2:]:
                    # argVector, body = exp.positional[0], exp.positional[1:]
                    fn = fn.register(Function(exp.positional[0], exp.positional[1:], env, sexp.relational))
                return fn
        elif is_s and "do" == head.name:
            if len(sexp.positional) == 1: return None
            for e in sexp[1:-1]:
                eval_sexp(e, env)
            sexp = sexp[-1]
            continue # TCO
        elif is_s and "let" == head.name:
            # creates a new environment and binds the values to it successively (in order)
            # evaluates the body in an implicit do
            if len(sexp.positional) == 2: raise SimbaSyntaxError(f"\n\t`let` form is missing a body in:\n\t{print_sexp(sexp)}")
            vec = sexp.positional[1]
            env = Environment(outer = env)
            for i in range(0, len(vec), 2):
                env.set(vec[i], eval_sexp(vec[i+1], env))
            for exp in sexp[2:-1]:
                eval_sexp(exp, env)
            sexp = sexp[-1]
            continue # TCO
        elif is_s and "try" == head.name:
            if Keyword('catch') in sexp and Keyword('finally') in sexp:
                try:
                    res = None
                    for exp in sexp[1:]:
                        res = eval_sexp(exp, env)
                    return res
                except:
                    return eval_sexp(sexp['catch'], env)
                finally:
                    eval_sexp(sexp['finally'], env)
            elif Keyword('catch') in sexp:
                try:
                    res = None
                    for exp in sexp[1:]:
                        res = eval_sexp(exp, env)
                    return res
                except:
                    return eval_sexp(sexp['catch'], env)
            elif Keyword('finally') in sexp:
                try:
                    res = None
                    for exp in sexp[1:]:
                        res = eval_sexp(exp, env)
                    return res
                finally:
                    eval_sexp(sexp['finally'], env)
            else:
                res = None
                for exp in sexp[1:]:
                    res = eval_sexp(exp, env)
                return res
        elif is_s and "ns" == head.name:
            ns_name = sexp.positional[1].name                
            if ns_name not in loaded_ns:
                if env.ns.name == "":
                    # it means the execution is not currently happening in an environment:
                    # reuse the empty shell of an env here
                    env.ns.name = ns_name
                else:
                    # create new Namespace if it does not yet exist
                    env = Namespace(name = ns_name)
                # add it to the list of loaded namespaces
                loaded_ns[ns_name] = env
            # includes the base library and imports all classes from java.lang
            require(['base'], env.ns, include = True)
            # requires/refers other namespaces and vars as specified
            # dadi
            break
        elif is_s and "require" == head.name:
            require([sexp.positional[1].name], env.ns)
            break
        elif is_s and "include" == head.name:
            require([sexp.positional[1].name], env.ns, include=True)
            break
        elif is_s and "quasiquote" == head.name:
            sexp = quasiquote(sexp[1])
            continue # TCO
        elif is_s and "macroexpand" == head.name:
            if len(sexp.positional) != 2: return Exception("`macroexpand` expected 1 argument.")
            return macroexpand(sexp[1], env)
        else:
            ## Non-Special Forms ##
            ## evaluate the args ##
            evaledPArgs = [eval_sexp(e, env) for e in sexp.positional]
            evaledRArgs = {eval_sexp(key, env): \
                            eval_sexp(sexp.relational[key], env) \
                            for key in sexp.relational}
            matched = False
            if isinstance(fun := evaledPArgs[0], Multi):
                for tup in fun.method_table:
                    if signature_match(tup[0], evaledPArgs[1:], evaledRArgs, guard = tup[2]):
                        fun = evaledPArgs[0] = tup[1]
                        matched = True
                        break
                if not matched: raise MultipleDispatchException(f"No matching signature {evaledPArgs} for multimethod {sexp[0].name}")
            if isinstance(fun, MultiFn):
                # evaluate the dispatch function
                val = eval_sexp(SymbolicExpression(fun.dispatch_fn, *evaledPArgs[1:], **evaledRArgs), env)
                # if the value is missing from the dispatch table fall back to "default"
                if val not in fun.method_table:
                    try:
                        fun = fun.method_table["default"]
                    except KeyError:
                        raise MultipleDispatchException(f"Tried to fall back to the default method but no such method exists for {fun.name}.")
                else:
                    fun = fun.method_table[val]
            if isinstance(fun, Function):
                env = Environment(
                    outer = fun.bindings,
                    names = bind_fn_args(fun.args, evaledPArgs[1:], evaledRArgs) # args are the names of the env
                )
                # first check the guard if there is a guard and the fn has not already been selected from a multi
                if fun.guard and not matched and not eval_sexp(fun.guard, env):
                    raise FailingGuardException('The guard condition is failing for the function.')
                # if the body is empty return `nil`
                if len(fun.ast) == 0:
                    return None
                for e in fun.ast[:-1]: eval_sexp(e, env)
                sexp = fun.ast[-1]
                continue
            else:
                ## apply the head ##
                return fun(*evaledPArgs[1:], **evaledRArgs)


def namespaced_eval(ast_list, run_tests = False, main_namespace = "main"):
    """Adds namespace functionality to the Simba evaluation.
    
    Keeps track of multiple namespaces in the namespaces variable.
    At any given time, a namespace is 'current', the current namespace is held in the ns variable. This is inspired by common lisp.
    
    The evaluation starts in the 'main' namespace. Every time a namespace is referred, that namespace the function looks for it in its
    AST representation of the program, and evaluates it if it is the first time it is referred."""
    if run_tests:
        # get the test namespaces
        test_ns_list = helpers.return_test_ns(ast_list)
        # run them
        for ns in test_ns_list:
            env = Namespace()
            for e in ns:
                eval_sexp(e, env)
        return
    main = helpers.return_ns(main_namespace, ast_list)
    env = Namespace()
    for e in main:
        eval_sexp(e, env)

def main():
    """Serves as an entrypoint for the script."""
    global program_ast
    base_lib = read_text(sb, 'base.sb')
    # when called with no arguments, the command is a terminal repl
    if len(sys.argv) == 1:
        program_ast = helpers.read_string_eager(base_lib)
        simba_repl()
    # when called with the name of a file, will interpret that file
    elif len(sys.argv) > 1:
        parser = argparse.ArgumentParser(description="This is the Simba programming language tool. Call without arguments to start a command line repl, else the arguments specify the files to execute. At the REPL, press ALT+ENTER to submit.")
        parser.add_argument("files", metavar="FILES", nargs='*')
        parser.add_argument("--main", default="main", help="specify the namespace to start execution in (defaults to `main`)")
        parser.add_argument("--run-tests", action='store_const', const=True, default=False, help="execute the tests, that is all the namespaces whose name ends in `-test`")
        args = parser.parse_args()

        program_ast = helpers.read_files(_reader, args.files) + helpers.read_string_eager(base_lib)
        namespaced_eval(program_ast, args.run_tests, args.main)

if __name__ == "__main__":
    main()


from simba_types import *
# from antlr4.tree.Tree import ParseTreeWalker
# import dot_expr_grammar as grammar # imports simbaLexer, simbaParser
# from src.dot_expr_grammar.simbaLexer import simbaLexer
# from src.dot_expr_grammar.simbaParser import simbaParser
# from src.dot_expr_grammar.simbaListener import simbaListener
import itertools
# import antlr4

def addIndentationTokens(input):
    """ 
        Before lexing and parsing we must insert tokens that stand for
        the start and end of an indented block in the source file. 
        The indentation rules follow these from the official python 
        reference: <https://docs.python.org/3.3/reference/lexical_analysis.html#indentation>

        Importantly, this script only adds to the end of exiting lines.
        Hence it does not affect line and column numbers in error reporting. 

        Input and output are strings for now.
    """
    stack = [0]
    indent = ' > ' # ' 🐼 '
    dedent = ' < ' # ' 🐨 '
    new = []
    for line_number, line in enumerate(input.split('\n')):
        if line.isspace() or line == "":
            new.append('')
            continue
        leadingSpaces = sum(1 for _ in itertools.takewhile(str.isspace, line))
        if stack[-1] < leadingSpaces:
            stack.append(leadingSpaces)
            new[-1] = new[-1] + indent
        else:
            while stack[-1] > leadingSpaces:
                stack.pop() 
                if stack[-1] < leadingSpaces:
                    raise IndentationError(f'Line number {line_number} is not indented properly.')
                new[-1] = new[-1] + dedent
        new.append(line)
    # handle end of file
    for _ in stack[1:]:
        new[-1] = new[-1] + dedent
    output = '\n'.join(new)
    return output

def read_sexp(inputStream, environment):
    # print('=== Output ===')
    lexer  = simbaLexer(inputStream)
    stream = antlr4.CommonTokenStream(lexer)
    parser = simbaParser(stream)
    tree   = parser.start()
    # print(tree.toStringTree(recog=parser))

    # listener = antlr4.ParseTreeListener()
    # ParseTreeWalker.DEFAULT.walk(listener, tree)

    # printer = simbaPrintListener()
    # walker = ParseTreeWalker()
    # walker.walk(printer, tree)
    # return dataStructure

# ================================================================
#                              PRINTER
# ================================================================

def to_string(obj, indent = 0):
    def _escape(s): return s.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
    if type(obj) == SymbolicExpression:
        return obj.positional[0] + ". " + " ".join(map(lambda e: to_string(e), obj.positional[1:]))
    elif type(obj) == Vector:                                    
        return "[" + " ".join(map(lambda e: to_string(e), obj)) + "]"
    elif type(obj) == Map:
        ret = []
        for k in obj.keys():
            ret.extend((to_string(k), to_string(obj[k])))
        return "{" + " ".join(ret) + "}"
    elif type(obj) == str:
        return '"' + _escape(obj) + '"'
    elif type(obj) == None:
        return "nil"
    elif type(obj) == True:
        return "true"
    elif type(obj) == False:
        return "false"
    # elif type(obj) == Atom:
    #     return "(atom " + to_string(obj.val,_r) + ")"
    else:
        return obj.__str__()


# Module containing some of the base functions for simba.
# Date created: 28 January 2022
# Author: Jérémie Vaney

import functools as ft
import itertools as it
import operator as op
from math import prod
from types import LambdaType

import pyrsistent
from simba.simba_types import Map, Vector, SymbolicExpression, Symbol
from simba.simba_exceptions import SimbaException
import simba.helpers as helpers
import time
import toolz

import collections.abc


def sb_apply(fn, *args, **kwargs):
    """Function application. Conses all the arguments between the first and the last to the last.
    Also applies any keyword arguments to the function."""
    arglist = []
    for e in args[:-1]:
        arglist.append(e)
    arglist = arglist + [*args[-1]]
    return fn(*arglist, **kwargs)

def sb_add(x, y):
    if x is None:
        return y
    if y is None:
        return None
    return x + y

def sb_print(e):
    from simba import print_sexp
    print(print_sexp(e))

def sb_seq(seqable):
    """Returns an appropriate sequence representation of an object.
    If the sequable is empty, return nil."""
    if isinstance(seqable, collections.abc.Sequence):
        if len(seqable) == 0:
            return None
        return seqable
    elif isinstance(seqable, Map):
        if len(s := seqable.items()) == 0:
            return None
        return s
    elif isinstance(seqable, dict):
        return pyrsistent.pvector(seqable.items())
    elif isinstance(seqable, SymbolicExpression):
        return pyrsistent.pvector(seqable)
    elif seqable is None:
        return None
    else:
        raise SimbaException(f"Don't know how to make a seq from {type(seqable)}", None, None, None)

def sb_prepend_sexp(e, seq):
    """Defines a generic prepend function that should work on all sequence types.
    Returns an object of the same type as the second argument.
    If second arg is None, defaults to a SymbolicExpression"""
    if seq is None:
        return SymbolicExpression(e)
    # elif isinstance(seq, tuple):
    #     return (e,) + seq
    # elif isinstance(seq, list):
    #     return [e] + seq
    # elif isinstance(seq, SymbolicExpression):
    return SymbolicExpression(e) + SymbolicExpression(*seq)

def sb_generic_concat(a, b):
    """Python has no built-in way of concatenating sequences of different types.
    Howver, this is often needed in Simba to perform generic operations on sequences.
    This function covers the built-in Python sequences and always returns a sequence
    of the same type as the first value."""
    # TODO: also define a function that works on n parameters
    if b is None:
        return a
    if a is None:
        return b
    if type(a) == type(b):
        return a+b
    elif isinstance(a, tuple):
        return a + tuple(b)
    elif isinstance(a, list):
        return a + list(b)
    elif isinstance(a, SymbolicExpression):
        return a + SymbolicExpression(*b)

def sb_uniform_access(*attrs):
    """This implements the Uniform Access Principle.
    Coined in Object-Oriented Software Construction by Bertrand Meyer:
    'All services offered by a module should be available through a uniform notation, which does not betray whether they are implemented through storage or through computation.'
    - https://en.wikipedia.org/wiki/Uniform_access_principle
    - http://wiki.c2.com/?UniformAccessPrinciple"""
    obj = attrs[-1]
    attr = attrs[0]
    if isinstance(attr, Symbol):
        attr = attr.name
    args = attrs[1:-1]
    if hasattr(obj, attr):
        a = getattr(obj, attr)
        if callable(a) and not isinstance(a, LambdaType):
            return eval(f"obj.{attr}")(*args)
        else:
            return a
    raise SimbaException(f"{obj} has no attribute {attr}", None, None, None)

def throw(e): raise e

# define operators that are infix in Python
repl_env = {
    # predicates
    'instance?': isinstance, # maybe I should switch the order of arguments
    'macro?': lambda obj: True if obj.macro else False, # need the environment as a param

    # arithmetic
    '+': lambda *a: ft.reduce(sb_add, a),
    '-': lambda a, *s: a-sum(s) if s else -a,
    '*': lambda *a: prod(a),
    '/': lambda a, *b: a / prod(b),
    '%': lambda a, b: b % a,
    'pow': lambda a, b: a ** b,
    'floordiv': lambda a, b: a // b,

    # comparison
    '=': lambda a, b: a == b,
    '!=': lambda a, b: a != b,
    '>': lambda a, b: a > b,
    '<': lambda a, b: a < b,
    '<=': lambda a, b: a <= b,
    '>=': lambda a, b: a >= b,

    # identity
    'is': lambda a, b: a is b,
    'is-not': lambda a, b: a is not b,

    # logic
    'not': lambda a: not a,
    'and': lambda a, b: a and b,
    'or': lambda a, b: a or b,
    # 'nand':
    # 'xor':

    # membership
    'in': lambda a, b: a in b,
    'not-in': lambda a, b: a not in b,

    # bitwise
    '&': lambda a, b: a & b,
    '|': lambda a, b: a | b,
    '^': lambda a, b: a ^ b,
    '~': lambda a: ~a,
    '<<': lambda a, b: a << b,
    '>>': lambda a, b: a >> b,

    # types
    'sexp': SymbolicExpression,
    'symbol': Symbol,
    'Vector': Vector,
    'vector': pyrsistent.v,
    'tuple': tuple,
    't': lambda *t: t,
    'HashMap': Map,
    'hash-map': pyrsistent.pmap,
    'set': set,
    'exception': Exception,
    'type': type,
    'str': str,
    'bytes': bytes,

    # magic or dunder methods: https://rszalski.github.io/magicmethods/
    'new': lambda obj, *args, **kwargs: obj(*args, **kwargs),
    'del': lambda obj: obj.__del__(),
    'at': lambda i, obj: obj[i], # 'get-item'
    'set-at': lambda i, new, obj: obj.__setitem__(i, new),
    'del-at': lambda i, obj: obj.__delitem__(i),
    # I would like to have
    # - special syntax for get, set, del
    # - universal access principle, such that if attr doesn't exist, then tries to call method
    'get-attr': sb_uniform_access, # lambda attr, obj: obj.__getattribute__(attr),
    'set-attr': lambda attr, value, obj: obj.__delattr__(attr, value),
    'del-attr': lambda attr, obj: obj.__delattr__(attr),

    # sequences
    'count': toolz.count,
    'between': lambda low, up = None, seq = None: seq[low:up] if seq is not None else up[low:],
    'prepend-sexp': sb_prepend_sexp,
    'prepend': lambda *es: list(es[:-1] + tuple(es[-1])),
    # 'append': sb_append,
    'concat': sb_generic_concat, # lambda *lists: ft.reduce(op.add, lists), # this poses certain problems bc its the add operation
    'reverse': helpers.reverse,
    'range': range,

    # collections

    # IO
    'print': (lambda *args: [sb_print(e) for e in args][0]),
    'prn': (lambda *args: [print(e, end = "") for e in args][0]),
    'throw': throw,

    # reflection
    # interop
    'get': lambda e, obj: obj[e],
    'date': time.time,
    'py-exec': helpers.exec_with_return,
}


import ast
from simbaTypes import *
import simba
import helpers
import reader

s = SymbolicExpression(Symbol('if'), True, 1, 0)

src = """\
(if true 1 0)
"""

readerObj = helpers.read_string_eager()



print(ast.dump(ast.parse('from y import x,y,z'), indent=4))

# https://python-ast-explorer.com/
# https://github.com/octoml/synr


import re
from simba.simba_types import (Symbol, Vector, Map, SymbolicExpression, Keyword, Namespace)
import pyrsistent as p
import pvectorc

class Blank(Exception): pass

class SexpReader():
    """Implementation of a recursive descent parser for s-expressions.
    Reads tokens one by one."""
    def __init__(self, text, position=0):
        self.tokens = tokenize(text)
        self.position = position

    def next(self):
        "Advance token position by one."
        self.position += 1
        return self.tokens[self.position-1]

    def peek(self):
        if len(self.tokens) > self.position:
            return self.tokens[self.position]
        else:
            return None

    def read_form(self):
        if self.position > len(self.tokens)-1: raise Exception("Token out of bounds exception")
        return read_form(self)

    def read_forms(self):
        ast = []
        while self.position < len(self.tokens):
            ast.append(read_form(self))

def _unescape(s):
    return s.replace('\\\\', '\u029e').replace('\\"', '"').replace('\\n', '\n').replace('\u029e', '\\')

def tokenize(str: str) -> list[str]:
        tre = re.compile(r"""[\s]*(~@|[\[\]{}()'`~^@]|"(?:[\\].|[^\\"])*"?|;.*|[^\s\[\]{}()'"`@,;]+)""");
        return [t for t in re.findall(tre, str) if t[0] != ';']

def read_atom(reader):
    int_re = re.compile(r"-?[0-9]+$")
    float_re = re.compile(r"-?[0-9][0-9.]*$")
    string_re = re.compile(r'"(?:[\\].|[^\\"])*"')
    token = reader.next()
    if re.match(int_re, token):     return int(token)
    elif re.match(float_re, token): return float(token)
    elif re.match(string_re, token):return _unescape(token[1:-1])
    elif token[0] == '"':           raise Exception("expected '\"', got EOF")
    elif token[0] == ':':           return Keyword(token[1:])
    elif token[0] == '.':           return token[1:]
    elif token == "nil":            return None
    elif token == "true":           return True
    elif token == "false":          return False
    else:                           return Symbol(token) # here returns symbol

def read_sequence(reader, start='(', end=')'):
    token = reader.next()
    if token != start: raise Exception("expected '" + start + "'")

    arguments = []
    token = reader.peek()
    while token != end:
        if not token: raise Exception("expected '" + end + "', got EOF")
        if token[0] == '.':
            # in the midst of this if clause
            arguments.append(Symbol('get-attr'))
            arguments.append(read_form(reader))
        arguments.append(read_form(reader))
        token = reader.peek()
    reader.next()
    # print(arguments)
    return arguments

def readMap(reader):
    lst = read_sequence(reader, '{', '}')
    m = {}
    for name, value in zip(lst[0::2], lst[1::2]):
        if isinstance(value, Keyword):
            raise Exception("A map can only have keywords as keys.")
        if isinstance(name, Keyword):
            name = str(name)
        m[name] = value
    return p.pmap(m)

def readSymbolicExpression(reader):
    return SymbolicExpression(*read_sequence(reader, '(', ')'))

def readVector(reader):
    return p.pvector(read_sequence(reader, '[', ']'))

def read_form(reader):
    "Reads a single Simba form. Raises an exception for unmatched parens."
    token = reader.peek()
    # reader macros/transforms
    if token[0] == ';':
        reader.next()
        return None
    elif token == '\'':
        reader.next()
        return SymbolicExpression(Symbol('quote'), read_form(reader))
    elif token == '`':
        reader.next()
        return SymbolicExpression(Symbol('quasiquote'), read_form(reader))
    elif token == '~':
        reader.next()
        return SymbolicExpression(Symbol('unquote'), read_form(reader))
    elif token == '~@':
        reader.next()
        return SymbolicExpression(Symbol('splice-unquote'), read_form(reader))
    # elif token == '^':
    #     reader.next()
    #     meta = read_form(reader)
    #     return SymbolicExpression(Symbol('with-meta'), read_form(reader), meta)
    elif token == '@':
        reader.next()
        return SymbolicExpression(Symbol('deref'), read_form(reader))
    # list
    elif token == ')': raise Exception("unexpected ')'")
    elif token == '(': return readSymbolicExpression(reader)
    # vector
    elif token == ']': raise Exception("unexpected ']'")
    elif token == '[': return readVector(reader);
    # map
    elif token == '}': raise Exception("unexpected '}'")
    elif token == '{': return readMap(reader);
    # atom
    else:              return read_atom(reader)

# def read_str(str):
#     """read_str tokenizes the input string and returns a Simba data structures"""
#     tokens = tokenize(str)
#     # if len(tokens) == 0: raise Blank("Blank Line")
#     return read_form(SexpReader(tokens))

# =============================================================
#                          PRINTER
# =============================================================

def to_string(obj, indent=0, lb=True) -> str:
    def _escape(s): return s.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
    if lb:
        start = '\n'
    else:
        start = ''
    if isinstance(obj, SymbolicExpression):
        if obj.positional and obj.relational:
            return start + " "*indent + "(" + " ".join(to_string(e,indent=indent+2) for e in obj.positional) + " " + " ".join(to_string(Keyword(key), indent=indent+2) + " " + to_string(obj.relational[key], indent=indent+2) for key in obj.relational) + ")"
        elif obj.relational:
            return "(" + " ".join(to_string(Keyword(key), indent=indent, lb=lb) + " " + to_string(obj.relational[key], indent=indent+2) for key in obj.relational) + ")"
        else: return start + " "*indent + "(" + " ".join(to_string(e, indent=indent+2) for e in obj.positional) + ")"
    elif isinstance(obj, Vector):
        return "[" + " ".join(to_string(e) for e in obj) + "]"
    elif isinstance(obj, Map):
        return "{" + " ".join(to_string(k, indent=indent, lb=lb) + " " + to_string(obj[k], indent=indent, lb=lb) + "\n" for k in obj) + "}"
    elif isinstance(obj, Keyword):
        return start + " "*indent + ':' + str(obj)
    elif isinstance(obj, str):
        # if len(obj) > 0 and obj[0] == '\u029e':
        #     return ':' + obj[1:]
        return '"' + _escape(obj) + '"'
    elif obj is None:
        return "nil"
    elif obj is True:
        return "true"
    elif obj is False:
        return "false"
    elif isinstance(obj, Namespace):
        return f"#namespace[{obj.name}]"
    else:
        # print(type(obj))
        return obj.__str__()


import os
from simba.simba_exceptions import MultipleNamespacesError, NoNamespaceError
import simba.sexprs_reader_printer as sexprs_reader_printer

def reverse(lst):
    return [ele for ele in reversed(lst)]

def read_files(reader, files):
    """Opens, reads, and closes the files in order.
    If the filename points to a directory, then read all the files in that directory.
    (and all the files in all the subdirectories).
    Returns a Simba Symbolic Expression."""
    ast_list = []
    for file in files:
        if os.path.isdir(file):
            for name in os.listdir(file):
                if file[-1] == '/':
                    ast_list += read_files(reader, [file + name])
                else:
                    ast_list += read_files(reader, ['/'.join([file, name])])
        else:
            with open(file, 'r') as f:
                if file[-3:] == ".sb":    
                    program = f.read()
                    readerObj = reader.SexpReader(program)
                    while readerObj.position < len(readerObj.tokens):
                        e = readerObj.read_form()
                        ast_list.append(e)
    return ast_list

def read_files_form_by_form(reader, procedure, files):
    """Same as read_files except that it reads the forms one by one and it can take a procedure to execute on each form.
    This function allows one to perform an operation on the data in one pass, unlike read_files.
    Furthermore, it is takes roughly 'constant' space where read_files takes linear space."""
    for file in files:
        with open(file, 'r') as f:
            program = f.read()
            readerObj = reader.SexpReader(program)
            while readerObj.position < len(readerObj.tokens):
                e = readerObj.read_form()
                procedure(e)

def read_string_eager(string, reader = sexprs_reader_printer):
    ast_list = []
    readerObj = reader.SexpReader(string)
    while readerObj.position < len(readerObj.tokens):
        e = readerObj.read_form()
        ast_list.append(e)
    return ast_list

def read_string_form_by_form(reader, procedure, string):
    """Similar to read_files_form_by_form, useful for the REPL."""
    readerObj = reader.SexpReader(string)
    while readerObj.position < len(readerObj.tokens):
        e = readerObj.read_form()
        procedure(e)

def find_forms(startswith, ast_list):
    """Finds the forms that begin with the elements in the startwith argument."""
    results = []
    for f in ast_list:
        is_match = True
        for i, val in enumerate(f[:len(startswith)]):
            if val.name != startswith[i]:
                is_match = False
        if is_match: results.append(f)
    return results

def find_test_ns(ast_list):
    """Finds all the test namespaces in the ast, which are all the namespaces that have a name ending in '-test'."""
    results = []
    for f in ast_list:
        if f[0].name == "ns" and f[1].name[-5:] == "-test":
            results.append(f)
    return results

def find_ns(name, ast_list):
    """Finds the namespace of a given name in the AST."""
    res = find_forms(["ns", name], ast_list)
    if len(res) > 1:
        raise Exception(f"There is more than one `{name}` namespace.")
    if len(res) == 0:
        raise Exception(f"No `{name}` namespace was found.")
    return res[0]

def return_ns(name, ast_list):
    from simba.simba_types import Symbol, SymbolicExpression
    res = []
    is_found = False
    adding = False
    for exp in ast_list:
        cond = isinstance(exp, SymbolicExpression)
        if cond and exp[0] == Symbol("ns") and exp[1] == Symbol(name):
            if is_found: raise MultipleNamespacesError(f"Multiple namespaces of the name {name} were found.")
            is_found = True
            adding = True
        elif cond and exp[0] == Symbol("ns"):
            adding = False
        if adding:
            res.append(exp)
    if not is_found: raise NoNamespaceError(f"No namespace of the name {name} was found.")
    return res

def return_test_ns(ast_list):
    from simba.simba_types import Symbol, SymbolicExpression
    res = []
    is_test = False
    is_found = False
    for exp in ast_list:
        cond = isinstance(exp, SymbolicExpression)
        if cond and exp[0] == Symbol("ns") and isinstance(exp[1], Symbol) and exp[1].name[-5:] == "-test":
            res.append([])
            is_test = True
            is_found = True
        elif cond and exp[0] == Symbol("ns"):
            is_test = False
        if is_test:
            res[-1].append(exp)
    if not is_found: raise NoNamespaceError(f"No test namespace was found.")
    return res

def get_base_namespace(reader):
    return read_files(reader, ['libraries/base.sb'])

#== Exec except that it returns the value ==#
# Taken from https://stackoverflow.com/questions/33409207/how-to-return-value-from-exec-in-function

import ast
import copy
def convertExpr2Expression(Expr):
        Expr.lineno = 0
        Expr.col_offset = 0
        result = ast.Expression(Expr.value, lineno=0, col_offset = 0)

        return result
def exec_with_return(code):
    code_ast = ast.parse(code)

    init_ast = copy.deepcopy(code_ast)
    init_ast.body = code_ast.body[:-1]

    last_ast = copy.deepcopy(code_ast)
    last_ast.body = code_ast.body[-1:]

    exec(compile(init_ast, "<ast>", "exec"), globals())
    if type(last_ast.body[0]) == ast.Expr:
        return eval(compile(convertExpr2Expression(last_ast.body[0]), "<ast>", "eval"),globals())
    else:
        exec(compile(last_ast, "<ast>", "exec"),globals())


class SimbaException(Exception): pass

class SimbaSyntaxError(SimbaException): pass

class IllegalArgumentException(SimbaException): pass
class MultipleDispatchException(SimbaException): pass
class MultiMethodException(SimbaException): pass
class FailingGuardException(SimbaException): pass

class UnresolvedSymbolError(SimbaException): pass
class MultipleNamespacesError(SimbaException): pass
class NoNamespaceError(SimbaException): pass
class ImmutableBindingException(SimbaException): pass